local Sackville = LibStub("AceAddon-3.0"):GetAddon("Sackville", "AceConsole-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Sackville")

Sackville.Bag = {}
Sackville.Bag.__index = Sackville.Bag

local _

function Sackville.Bag:New(bagName, bagConfig)
    Sackville:Debug(1, "bag", "Adding new virtual bag: %s", bagName)
    local self = {}
    setmetatable(self, Sackville.Bag)

    local bagDefaults = Sackville.Options.bag
    self.name = bagName
    self.displayOrder = bagConfig.displayOrder
    self.bagType = bagConfig.bagType
    self.defaultPriority = bagConfig.defaultPriority or 1
    self.defaultSortCriteria = bagConfig.sortCriteria or {}
    self.minimized = false

    self.frame = self:CreateBagFrame(self.name, self.bagType, true)
    self.min_width = self.frame:GetWidth()
    self.max_width = bagDefaults.max_width  
    self.sections = {}
    return self
end

function Sackville.Bag:AddSection(sectionName, sectionOptions)
    local bagDefaults = Sackville.Options.bag
    local section = Sackville.Section:New(self, sectionName, sectionOptions)
    section:SetMaxWidth(self.max_width - bagDefaults.margin_left 
        - bagDefaults.margin_right)
    self.sections[section:GetName()] = section
end

function Sackville.Bag:RemoveSection(sectionName)
    self.sections[sectionName] = nil
end

function Sackville.Bag:CreateBagFrame(bagName, bagType, moveable)
    Sackville:Debug(1, "bag", "CreateBagFrame: %s", bagName)
    local frameName = "SackvilleVB_" .. bagName
    local f = _G[frameName] or CreateFrame('Frame', frameName, UIParent)
    local bagDefaults = Sackville.Options.bag
    f:Hide()

    if bagType == PLAYER then
        f:SetBackdrop(bagDefaults.backdrop)
    elseif bagType == L["Bank"] then
        f:SetBackdrop(bagDefaults.backdrop_bank)
    end
    f:SetHeight(bagDefaults.min_height)

    f:SetFrameStrata(bagDefaults.strata)
    
    if (moveable) then
        f:SetMovable(true)
        f:EnableMouse(true)
        f:RegisterForDrag('LeftButton')
        f:SetScript('OnMouseDown', function(f, button)
         if button ~= 'LeftButton' then return end
         f.pOrigin = f.pOrigin or {}
         f.pOrigin.point, f.pOrigin.anchor, f.pOrigin.relPoint, f.pOrigin.x, f.pOrigin.y
            = f:GetPoint(1)
         f:StartMoving()
         f.pStart = f.pStart or {}
         f.pStart.x, f.pStart.y = select(4, f:GetPoint(1))
        end)
        f:SetScript('OnMouseUp', function(f, button)
         if button ~= 'LeftButton' then return end
         f.pEnd = f.pEnd or {}
         f.pEnd.x, f.pEnd.y = select(4, f:GetPoint(1))
         f:StopMovingOrSizing()
         f:ClearAllPoints()
         local x = f.pOrigin.x + (f.pEnd.x - f.pStart.x)
         local y = f.pOrigin.y + (f.pEnd.y - f.pStart.y)
         f:SetPoint(f.pOrigin.point, f.pOrigin.anchor, f.pOrigin.relPoint, x, y)
        end)
    end


    local titleFrame =    _G[frameName .. "_TitleFrame"] or 
        CreateFrame("Frame", frameName .. "_TitleFrame", f)
    titleFrame:ClearAllPoints()
    titleFrame:SetPoint("TOPLEFT", 0, 0)
    titleFrame:SetHeight(bagDefaults.title_height)
    -- titleFrame:SetBackdrop(bagDefaults.backdrop_title)
    titleFrame:Show()

    local closebtn = _G[frameName .. "_CloseButton"] or 
        CreateFrame("Button", frameName .. "_CloseButton",
        titleFrame, "SackvilleBagCloseButton")
    closebtn:SetPoint("RIGHT", 0, 0)
    closebtn:Show()
    
    local minbtn = _G[frameName .. "_MinimizeButton"] or
        CreateFrame("Button", frameName .. "_MinimizeButton",
        titleFrame, "SackvilleBagMinimizeButton")
    minbtn:SetPoint("RIGHT", -20, 0)
    minbtn:Show()
    minbtn.bag = self
    
    local titleFontString = _G[frameName .. "_TitleText"] or 
        titleFrame:CreateFontString(frameName .. "_TitleText")
    titleFontString:SetFontObject("GameFontNormalLarge")
    titleFontString:SetText(bagName)
    titleFontString:ClearAllPoints()
    titleFontString:SetPoint("LEFT", 12, 0)
    titleFrame.text = titleFontString
    -- See http://www.wowwiki.com/API_FontInstance_SetFontObject

    titleFrame:SetWidth(titleFontString:GetStringWidth() + Sackville.Options.TITLE_EXTRA_WIDTH)
    f:SetWidth(titleFrame:GetWidth())

    f.titleFrame = titleFrame
    
    return f
end

function Sackville.Bag:CreateQuickPanel()
    local qpanel = Sackville.Bag:CreateBagFrame(Sackville.version, PLAYER, true)
    moneyFontString = qpanel:CreateFontString("Sackville_money")
    moneyFontString:SetFontObject("GameFontWhite")
    moneyFontString:SetText(Sackville.GetMoneyString())
    moneyFontString:ClearAllPoints()
    -- moneyFontString:SetPoint("BOTTOMLEFT", 30, 10)
    moneyFontString:SetPoint("BOTTOMLEFT", 30, 5)
    
    local editSearch = _G["Sackville_qpanelSearch"] or
        CreateFrame("EditBox", "Sackville_qpanelSearch", qpanel, "SearchBoxTemplate")
    editSearch:SetPoint("TOPLEFT", 15, -30)
    editSearch:SetHeight(15)
    editSearch:SetWidth(qpanel:GetWidth() - 30)
    editSearch:SetScript("OnTextChanged", function (text)
        Sackville:SearchBoxUpdate(text) end)

    qpanel.money = moneyFontString
    qpanel:SetHeight(70)
    w = qpanel.money:GetStringWidth() + Sackville.Options.TITLE_EXTRA_WIDTH
    qpanel:SetWidth(w)
    qpanel.titleFrame:SetWidth(w)
    return qpanel
end

function Sackville.Bag:MinimizeClicked()  
    if self.name then
        self.minimized = not self.minimized
        self:LayoutItems()
        Sackville:LayoutBags(self.displayOrder)
    end
end

function Sackville.Bag:CreateBagControls()
    local controlFrame = AceGUI:Create("SimpleGroup")
    controlFrame:SetLayout("Flow")
    
    local btnMin = AceGUI:Create("SackvilleButton")
    btnMin:SetAutoWidth(true)
    btnMin:SetText("_")
    controlFrame:AddChild(btnMin)
    
    local btnClose = AceGUI:Create("SackvilleButton")
    btnClose:SetAutoWidth(true)
    btnClose:SetText("X")
    controlFrame:AddChild(btnClose)
    
    return controlFrame
end

function Sackville.Bag:LayoutItems()
    Sackville:Debug(2, "bag", "LayoutItems for bag: %s", self.name)
    local bagDefaults = Sackville.Options.bag
    local sectionDefaults = Sackville.Options.section
    local BUTTON_SIZE = Sackville.Options.BUTTON_SIZE
    local state = {
        x_pos = bagDefaults.margin_left,
        y_pos = -bagDefaults.margin_top - Sackville.Options.section.margin_top,
        row = 0,
        col = 0,
        widest = 0
    }

    local titleInfo = {}

    for _, section in self:SectionIterator() do
        local sectionTitle = self:MakeSectionTitle(section.name)
        if section:HasItems() or section.alwaysShow then
        -- if section:HasItems() then
            titleInfo.start_x = state.x_pos
            titleInfo.start_y = state.y_pos
            titleInfo.end_x = titleInfo.start_x + BUTTON_SIZE
            
            Sackville:Debug(2, "bag", "Laying out section %s which has items.", section.name)
            for _, item in section:ItemIterator() do
                if self.minimized then
                    item:Hide()
                else
                    if titleInfo.start_y == state.y_pos then 
                        titleInfo.end_x = state.x_pos
                    end
                    state = self:PositionItem(item, state)
                    Sackville:Debug(3, "bag", "state => x: %d, y: %d, row: %d, col: %d",
                        state.x_pos, state.y_pos, state.row, state.col)
                end
            end
            local emptyItem = section:UpdateEmptySlot(self.bagType)
            if section.showEmptySlot and emptyItem then
                emptyItem:SetParent(self.frame)
                state = self:PositionItem(emptyItem, state)
            end
            if state.col ~= 0 then 
                state.x_pos = state.x_pos + sectionDefaults.padding_right
            end

            if self.minimized then
                sectionTitle:Hide()
            else
                sectionTitle:Show()

                sectionTitle:SetPoint("TOPLEFT", titleInfo.start_x, 
                    titleInfo.start_y + Sackville.Options.section.margin_top)

                titleWidth = titleInfo.end_x - titleInfo.start_x + BUTTON_SIZE + 
                    (sectionDefaults.padding_right / 2)
                sectionTitle:SetWidth(titleWidth)
                Sackville:Debug(3, "bag", "Made title %s, x:%d, y:%d, w:%d, frame:%s",
                    sectionTitle:GetText(), titleInfo.start_x, 
                    titleInfo.start_y + Sackville.Options.section.margin_top,
                    titleWidth, sectionTitle:GetName())
            end
        else
            sectionTitle:Hide()
        end
    end
    
    if state.row > 0 then 
        self.frame:SetWidth(self.max_width)
    else
        self.frame:SetWidth(math.max(self.min_width, state.widest))
    end
        
    if self.minimized then
        self.frame:SetHeight(self.frame.titleFrame:GetHeight())
    else
        if state.col == 0 then
            self.frame:SetHeight(math.abs(state.y_pos) + bagDefaults.margin_bottom )
        else
            self.frame:SetHeight(math.abs(state.y_pos) + BUTTON_SIZE + bagDefaults.margin_bottom )
        end
    end
    self.frame.titleFrame:SetWidth(self.frame:GetWidth())

    return true -- FIXME: return whether or not we need to be repositioned as a bag.
end

function Sackville.Bag:PositionItem(item, state)
    local BUTTON_SIZE = Sackville.Options.BUTTON_SIZE
    local bagDefaults = Sackville.Options.bag
    local margin_top = Sackville.Options.section.margin_top
      
    item:SetParent(self.frame)
    item:SetPoint("TOPLEFT", state.x_pos, state.y_pos)
    item:Show()
    
    state.x_pos = state.x_pos + BUTTON_SIZE
    if state.x_pos > state.widest then 
        state.widest = state.x_pos + bagDefaults.margin_left 
    end
    
    if (state.x_pos + BUTTON_SIZE + bagDefaults.margin_right > self.max_width) then
        widest = self.max_width
        state.row = state.row + 1
        state.col = 0
        state.x_pos = bagDefaults.margin_left
        state.y_pos = state.y_pos - margin_top - BUTTON_SIZE
        -- state.y_pos = state.y_pos - BUTTON_SIZE
    else
        state.col = state.col + 1
    end
    
    return state
end

function Sackville.Bag:MakeSectionTitle(sectionName)
    local basename = string.format("Sackville_%s_%s", self.name, sectionName)

    local text = _G[basename .. "_TitleText"] or 
        self.frame:CreateFontString(basename .. "_TitleText")
    text:SetParent(self.frame)
    text:SetFontObject("GameFontWhite")
    text:SetText(sectionName)
    text:ClearAllPoints()
    text:SetHeight(Sackville.Options.section.title_height)
    text:SetJustifyH("LEFT")
    
    return text
end

function Sackville.Bag:SectionIterator()
    sectionSortCritera = { { key = "displayOrder", comparison = "<" } }
    return Sackville.Util.kpairs(self.sections, sectionSortCritera)
end

function Sackville.Bag:Toggle()
    if (self:GetFrame():IsVisible()) then
        self:Hide()
    else
        self:Show()
    end
end

function Sackville.Bag:Show()
    self:LayoutItems()
    self:GetFrame():Show()
end

function Sackville.Bag:Hide()
    self:GetFrame():Hide()
end

function Sackville.Bag:HideSections()
    for _, section in self:SectionIterator() do
        section:Hide()
    end
end

function Sackville.Bag:ShowSections()
    for _, section in self:SectionIterator() do
        section:Show()
    end
end

function Sackville.Bag:HighlightSearchItems(text)
    for _, section in self:SectionIterator() do
        section:HighlightSearchItems(text)
    end
end

function Sackville.Bag:GetFrame()
    return self.frame
end

